/* ###*B*###
 * ERIKA Enterprise - a tiny RTOS for small microcontrollers
 *
 * Copyright (C) 2009-2010  Evidence Srl
 *
 * This file is part of ERIKA Enterprise.
 *
 * ERIKA Enterprise is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation,
 * (with a special exception described below).
 *
 * Linking this code statically or dynamically with other modules is
 * making a combined work based on this code.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this code with independent modules to produce an
 * executable, regardless of the license terms of these independent
 * modules, and to copy and distribute the resulting executable under
 * terms of your choice, provided that you also meet, for each linked
 * independent module, the terms and conditions of the license of that
 * module.  An independent module is a module which is not derived from
 * or based on this library.  If you modify this code, you may extend
 * this exception to your version of the code, but you are not
 * obligated to do so.  If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * ERIKA Enterprise is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License version 2 for more details.
 *
 * You should have received a copy of the GNU General Public License
 * version 2 along with ERIKA Enterprise; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 * ###*E*### */
/*
 * Another minimal EE demo that demonstrates how to work with the MULTISTACK
 * mode.
 * In this demo two task are activated inside an ISR and work with different
 * stacks.
 * With this demo you can test the preemption mechanism with stack change
 * starting from an interrupt.
 *
 *	Author: 2014 Martin Hoffmann, FAU Erlangen
 */

#include "stdlib.h"
#include "ee.h"

// #define TEST_IRQ 42
// 
// ISR(TEST_IRQ) {
//     (void)ctx;
//     ActivateTask(Task1);
//     ActivateTask(Task2);
//     counter_isr++;
//     printf("----------Test ISR (irq %d), cnt: %d----------\n", vector, counter_isr);
// 
//     EE_lapic_send_eoi();
// }
// 
// 
// 
// TASK(Task1)
// {
//     counter_task1++;
// 
//     printf("\tTask1 started. cnt: %d\n", counter_task1);
//     printf("\tTask1 finished.\n");
// }
// 
// TASK(Task2)
// {
//     counter_task2++;
// 
//     printf("\t\tTask2 started. cnt: %d\n", counter_task2);
//     printf("\t\tTask2 finished.\n");
// }
// 


//extern "C" volatile uint32_t random_source =0 ;
DeclareTask(SignalGatherInitiateTask);
DeclareTask(SignalGatherFinishedTask);
DeclareTask(SignalGatherTimeoutTask);
DeclareTask(SignalProcessingActuateTask);
DeclareTask(SignalProcessingAttitudeTask);
DeclareTask(ActuateTask);
DeclareTask(FlightControlAttitudeTask);
DeclareTask(FlightControlActuateTask);
DeclareTask(MavlinkSendTask);
DeclareTask(CopterControlTask);
DeclareTask(MavlinkRecvHandler);

// DeclareResource(SPIBus);

// DeclareAlarm(CopterControlWatchdogAlarm);

//! test finish function
__NEVER_INLINE__ void test_finish();

//! FAIL* checkpoint/tracing memory location
__attribute__((weak)) volatile unsigned int fail_trace = 0;

__ALWAYS_INLINE__ void test_start()
{
#ifndef FAIL
    printf("-- Trace start\n");
#endif
}

__NEVER_INLINE__ void test_finish()
{
    #ifndef FAIL
    printf("-- Trace done\n");
    #else
    // to prevent compiler from optimizing the call
    asm volatile ("nop");
    #endif
}


__ALWAYS_INLINE__ void test_trace(unsigned int val)\
{
    fail_trace = val;

    #ifndef FAIL
    static int i = 1;

    printf("T %d: %x\n", i++, val);
    #endif
}

int round;

#define MavlinkRecvHandler 42

TASK(SignalGatherInitiateTask) {
	if (round == 9) {
		test_trace(0x00);
		test_finish();
        EE_hal_shutdown_system();
	}
	test_trace(0x01);
	GetResource(SPIBus);
	test_trace(0x02);
	if ((round % 2) == 0) {
		ActivateTask(SignalGatherTimeoutTask);
		test_trace(0x03);
	} else {
		ActivateTask(SignalGatherFinishedTask);
		test_trace(0x04);
	}
	round ++;
	test_trace(0x05);
	ReleaseResource(SPIBus);
	test_trace(0x06);
	TerminateTask();
    test_trace(0x07);
}

TASK(SignalGatherFinishedTask) {
	test_trace(0x11);
	ActivateTask(SignalProcessingAttitudeTask);
	test_trace(0x12);
	ActivateTask(SignalProcessingActuateTask);
	test_trace(0x13);
	TerminateTask();
}

TASK(SignalGatherTimeoutTask) {
	test_trace(0x21);
	GetResource(SPIBus);
	test_trace(0x22);
	ReleaseResource(SPIBus);
	test_trace(0x23);
    ChainTask(SignalGatherFinishedTask);
}

volatile int calculate;

TASK(SignalProcessingActuateTask) {
	test_trace(0x31);
	// for (calculate = 0; calculate < 200; calculate++);
	test_trace(0x30 | (calculate & 0xf));
	TerminateTask();
}

TASK(SignalProcessingAttitudeTask) {
	test_trace(0x41);
	// for (calculate = 0; calculate < 200; calculate++);
	test_trace(0x40 | (calculate & 0xf));
	TerminateTask();
}

TASK(FlightControlTask) {
	test_trace(0x51);
	ActivateTask(FlightControlAttitudeTask);
	test_trace(0x52);
	ActivateTask(FlightControlActuateTask);
	test_trace(0x53);
	ActivateTask(MavlinkSendTask);
	test_trace(0x54);
	TerminateTask();
}

TASK(FlightControlAttitudeTask) {
	test_trace(0x61);
	TerminateTask();
}

TASK(FlightControlActuateTask) {
	test_trace(0x71);
	TerminateTask();
}

TASK(MavlinkSendTask) {
	test_trace(0x81);
	GetResource(SPIBus);
	test_trace(0x82);
	EE_lapic_trigger(MavlinkRecvHandler);
	ReleaseResource(SPIBus);
	test_trace(0x83);
	TerminateTask();
}

TASK(CopterControlTask) {
	test_trace(0x91);
	SuspendAllInterrupts();
	test_trace(0x92);
	ResumeAllInterrupts();
	test_trace(0x93);
	if (round < 5) {
		CancelAlarm(CopterControlWatchdogAlarm);
		test_trace(0x94);
		SetRelAlarm(CopterControlWatchdogAlarm, 100, 100);
	}
	test_trace(0x95);

	TerminateTask();
}

ISR(MavlinkRecvHandler) {
    test_trace(0xA1);
	ActivateTask(CopterControlTask);
	test_trace(0xA2);
    EE_lapic_send_eoi();
}
 
TASK(CopterControlWatchdogTask) {
	test_trace(0xB1);
    TerminateTask();
}

ISR(IRQ_PIT) {
    (void)ctx;
    // printf("Timer! (#%d)\n", vector);
    CounterTick(C1);
    EE_lapic_send_eoi();
}




/*
 * MAIN TASK
 */
int main(void)
{

    EE_uart_init(COM1);
    printf("Starting ERIKA\n");

    EE_pit_init();
    EE_pit_set_periodic(1000); // 1000 Hz

    SetRelAlarm(SignalGatherAlarm,  10, 30);
    SetRelAlarm(FlightControlAlarm, 13, 90);

    EE_hal_enableIRQ(); // enable interrupt

    //EE_lapic_trigger(TEST_IRQ);
    //while(counter_isr < 1);

    //EE_lapic_trigger(TEST_IRQ);
    //while(counter_isr < 2);

    StartOS(OSDEFAULTAPPMODE);

    /* now the background activities... */
    while(1)
    {
        asm volatile ("sti; hlt");
    }


    printf("Shutdown ERIKA\n");
    EE_hal_shutdown_system();
    __builtin_unreachable();
}
